"use strict"

import view from "./view.js";

class UISurveyMatrixViewController extends HTMLElement {

  static get observedAttributes(){ return ["value"]; }

  constructor(model){
    super();
    this.model = model || {};

    this.state = {};
    this.state.connected = false;
    //this.state.lastResponseTime = performance.now();

    //Keeps reference of events with bindings (so we can remove them)
    //see: https://stackoverflow.com/questions/11565471/removing-event-listener-which-was-added-with-bind
    this.event = {};
    this.event.onResponse = this._onResponse.bind(this);
    this.event.complete = new CustomEvent("complete", {detail: this.model});
    this.event.change = new CustomEvent("change", {detail: this.model});
    this.event.update = new CustomEvent("update", {detail: this.model});

    this.shadowRoot = this.attachShadow({mode: "closed"});
    this.shadowRoot.appendChild(view.content.cloneNode(true));

    this.view = {};
  }

  //Fires when the element is inserted into the DOM. It's a good place to set
  //the initial role, tabindex, internal state, and install event listeners.
  //
  //NOTE: A user may set a property on an instance of an element, before its
  //prototype has been connected to this class. The _upgradeProperty() method
  //will check for any instance properties and run them through the proper
  //class setters.
  connectedCallback() {
    //Wire views here
    this.view.container = this.shadowRoot.querySelector(".container");
    this.view.topContainer = this.shadowRoot.querySelector("#topContainer");
    this.view.instructions = this.shadowRoot.querySelector("#instructions");
    this.view.headersContainer = this.shadowRoot.querySelector("#headersContainer");
    this.view.questionsContainer = this.shadowRoot.querySelector("#questionsContainer");
    this.view.errorMessage = this.shadowRoot.querySelector("#errorMessage");

    //Reference events with bindings
    // Let the element we're live
    this.state.connected = true;
    this.startTimeStamp = performance.now();

    this._updateView();
  }

  attributeChangedCallback(attrName, oldVal, newVal) {
    switch(attrName){
      case "value":
        if(newVal !== this.value){ this.value = newVal; }
        break;
      default:
        throw new Error(`Attribute ${attrName} is not handled, you should probably do that`);
    }
  }

  get shadowRoot(){return this._shadowRoot;}
  set shadowRoot(value){ this._shadowRoot = value}

  get data(){
    if(!this._data){
      this._data = [];
      this.model.questions.forEach((question,index) => {
        const muggleIndex = index+1;
        const datum = {}
        datum.question = muggleIndex;
        datum.name = `${this.model.name}-Q${muggleIndex}`;
        datum.rt = null;
        datum.response = null;
        this._data.push(datum);
      })
    }

    return this._data
  }

  get value(){ return this.model; }
  set value(value){
    //Check if attribute matches property value, Sync the property with the
    //attribute if they do not, skip this step if already sync
    if(this.getAttribute("value") !== value){
      //By setting the attribute, the attributeChangedCallback() function is
      //called, which inturn calls this setter again.
      const stringValue = JSON.stringify(value);
      this.setAttribute("value", stringValue);
      //attributeChangeCallback() implicitly called
      return;
    }

    switch(value.constructor){
      case String:
        this.model = JSON.parse(value);
        break;
      case Object:
        this.model = value;
        break;
    }

    this.model.type = "ui-survey-matrix";
    this._updateView();
  }

  get hasInvalidResponse(){
    return !this.hasValidResponse;
  }

  get hasValidResponse(){
    let missingResponse = false;
    this.data.forEach(datum => {
      if(datum.response === null){
        missingResponse = true;
      }
    })
    return this.model.required? missingResponse === false : true;
  }

  _updateView(view) {
    //No point in rendering if there isn't a model source, or a view on screen
    if(!this.model || !this.model.headers || !this.state.connected){ return; }
    this.view.container.style.opacity = 1;
    const columnClass = `col-${Math.floor(12/this.model.headers.length)}`

    switch(view){

      default:
        if(!this.model.headers || !this.model.questions) return;
        this.view.instructions.innerHTML = this.model.instructions;
        this.view.headersContainer.innerHTML = "";
        this.view.questionsContainer.innerHTML = "";

        this.model.headers.forEach(header => {
          const headerView = this._getHeaderView(header, columnClass);
          this.view.headersContainer.appendChild(headerView);
        })

        // For each question, insert a row of items
        this.view.items = [];
        this.model.questions.forEach((question, questionIndex) => {
          // Create a view for each question, and start counting from 1, like muggles
          const muggleIndex = questionIndex+1;
          const questionView = this._getQuestionView(muggleIndex, question);
          this.view.questionsContainer.appendChild(questionView);
          // Create an item row
          const itemsContainer = this._getItemContainerView();
          questionView.appendChild(itemsContainer);
          // For each header, insert an item to itemsContainer
          this.model.headers.forEach( (headerModel, itemIndex) => {
            const muggleQuestionIndex = questionIndex + 1;
            const muggleItemIndex = itemIndex + 1;
            const itemContainer = this._getItemView(muggleQuestionIndex, muggleItemIndex, columnClass);
            const headerView = this.view.headersContainer.children[itemIndex];

            itemContainer.addEventListener("mouseover", () => { this._highlight(headerView, questionView, true); });
            itemContainer.addEventListener("mouseout", () => { this._highlight(headerView, questionView, false); });
            itemContainer.addEventListener("change", this.event.onResponse );
            itemsContainer.appendChild(itemContainer);

            const itemInput = itemContainer.querySelector("input")
            this.view.items.push(itemInput);
          })
        })
    }
  }

  //TODO:
  // - Handle end of page scrolling (infinite loop sice it cannot reach it)
  scrollTo(element) {
    const rect = element.getBoundingClientRect();
    const targetWithMargins = rect.y + 25; //margins
    if(targetWithMargins >= 0){
      scrollBy(0,11);
      requestAnimationFrame(()=>{this.scrollTo(element)})
    }
  }

  _onResponse(e){
    this.data.forEach((datum, index) => {
      const muggleIndex = index + 1;
      const responseIndex = parseInt(e.target.name);
      if(responseIndex === muggleIndex){
        datum.rt = e.timeStamp - this.startTimeStamp; //this.state.lastResponseTime;
        datum.response = parseInt(e.target.value);
      }
    })

    this.dispatchEvent(new CustomEvent("response", {detail: this.data}));
    this._checkIfCompleted();
  }

  scrollToY(y){
    const rect = this.getBoundingClientRect();
    const targetY = Math.abs(y);
    const currentY = Math.abs(rect.y);
    if(targetY>= currentY){
      scrollBy(0,2);
      requestAnimationFrame(()=>{this.scrollToY(targetY)});
    }
  }

  _checkIfCompleted(){
    let numResponses = 0;
    let numQuestions = this.model.questions.length;
    this.view.items.forEach(item => { numResponses += item.checked? 1 : 0; })
    const isComplete = numResponses === numQuestions;
    if(isComplete){
      this.dispatchEvent(new CustomEvent("complete", {detail: this.data}));
    }
  }

  _highlight(header, question, highlight){
    const fontWeight = highlight? 600 : 500;
    header.style.fontWeight = fontWeight;
  }

  _getItemContainerView(){
    return this._elementify(" <div class=\"row col-9\"> </div>");
  }

  _getItemView(name, value, numCols){
    return this._elementify(`
        <label class="radio-container ${numCols}">
            <input type="radio" name="${name}" value="${value}"/>
            <span class="checkmark"></span>
        </label> `);
  }

  _getHeaderView(innerHTML, numCols){
    return this._elementify(
      `<div class="${numCols} item text-center">
            ${innerHTML}
         </div>`);
  }

  _getQuestionView(index, innerHTML){
    return this._elementify(`
        <div class="question row no-gutters" data-index="${index}">
            <div class="col-3 question-text mt-3 mr-4">
                <p>${index}. ${innerHTML}</p>
            </div>
        </div>`);
  }

  _elementify(stringLiteral){
    const placeholder = document.createElement("div");
    placeholder.innerHTML = stringLiteral;
    return placeholder.querySelector(":first-child");
  }

  _removeEvents(){
    this.view.container.removeEventListener("click", this.event.click);
  }

  showValidationError(error){
    error = error || "Please answer all questions."
    this.view.errorMessage.innerText = error;
  }

  hideValidationError(){
    this.view.errorMessage.innerText = "";
  }

  disconnectedCallback() {
    this._removeEvents()
    this.state.connected = false;
  }
}

window.customElements.define("ui-survey-matrix", UISurveyMatrixViewController);
